import numpy as np
import fqsdfqsdfqsdfqsd.contracts as contracts
import fqsdfqsdfqsdfqsd.wallets as wallets


class PricingOracle:
    probability_of_default = 0.04
    loss_given_default = 0.5
    discount_rate = 0.05
    n_seconds_year = 31536000

    @classmethod
    def IT_price(cls):
        USDC_repayed = contracts.RepaymentContract.USDC_amount
        credit_outstanding_senior = cls.get_credit_outstanding_senior()
        USDC_deals_senior = cls.get_USDC_deals_senior()
        IT_in_circulation = cls.get_IT_in_circulation()

        if IT_in_circulation == 0:
            price = 1
        else:
            price = (USDC_repayed + USDC_deals_senior + credit_outstanding_senior) / IT_in_circulation

        return price

    @classmethod
    def TVL(cls):
        USDC_repayed = contracts.RepaymentContract.USDC_amount
        credit_outstanding_senior = cls.get_credit_outstanding_senior()
        return USDC_repayed + credit_outstanding_senior

    @classmethod
    def calculate_trailing_APY(cls, price_df):
        price_df['apy'] = ((price_df['IT price'] - price_df.shift(30)['IT price']) / 30) * 365
        price_df["apy"] = cls.hampel(price_df["apy"], 10, 0.3)

        return price_df['apy']

    @staticmethod
    def hampel(vals_orig, k=7, t0=3.0):
        '''
        vals: pandas series of values from which to remove outliers
        k: size of window (including the sample; 7 is equal to 3 on either side of value)
        '''

        # Make copy so original not edited
        vals = vals_orig.copy()

        # Hampel Filter
        L = 1.4826
        rolling_median = vals.rolling(window=k, center=True).median()
        MAD = lambda x: np.median(np.abs(x - np.median(x)))
        rolling_MAD = vals.rolling(window=k, center=True).apply(MAD)
        threshold = t0 * L * rolling_MAD
        difference = np.abs(vals - rolling_median)

        '''
        Perhaps a condition should be added here in the case that the threshold value
        is 0.0; maybe do not mark as outlier. MAD may be 0.0 without the original values
        being equal. See differences between MAD vs SDV.
        '''

        outlier_idx = difference > threshold
        vals[outlier_idx] = rolling_median[outlier_idx]
        return (vals)

    @classmethod
    def get_USDC_deals_senior(cls):
        USDC_deals_senior = 0
        deals = contracts.DealContract.get_instances()
        for deal in deals:
            USDC_deals_senior += deal.senior_tranche_current

        return USDC_deals_senior

    @classmethod
    def get_credit_outstanding(cls):
        credit_outstanding = 0
        deals = contracts.DealContract.get_instances()
        for deal in deals:
            credit_outstanding += deal.credit_outstanding

        return credit_outstanding

    @classmethod
    def get_credit_outstanding_senior(cls):
        credit_outstanding_senior = 0
        deals = contracts.DealContract.get_instances()
        for deal in deals:
            credit_outstanding_senior += deal.credit_outstanding * (deal.leverage_ratio / (deal.leverage_ratio + 1))

        return credit_outstanding_senior

    @classmethod
    def USDC_to_IT(cls, USDC_amount):
        return USDC_amount / cls.IT_price()

    @classmethod
    def IT_to_USDC(cls, IT_amount):
        return IT_amount * cls.IT_price()

    @classmethod
    def get_IT_in_circulation(cls):
        IT_in_circulation = 0
        investor_wallets = wallets.InvestorWallet.get_instances()
        for investor_wallet in investor_wallets:
            IT_in_circulation += investor_wallet.IT_balance

        return IT_in_circulation



from fqsdfqsdfqsdfqsd.wallets.wallet import Wallet
from fqsdfqsdfqsdfqsd.contracts import MatchContract

class InvestorWallet(Wallet):
    _instances = set()

    def __init__(self, USDC_balance=1000):
        super().__init__(USDC_balance=USDC_balance)
        self.UT_balance = None
        self._instances.add(self)

    @classmethod
    def get_instances(cls):
        return cls._instances

    @classmethod
    def del_instance(cls, instance):
        cls._instances.remove(instance)

    @classmethod
    def withdraw_IT(cls, IT_amount_to_withdraw, reserve):
        investor_wallets = cls.get_instances()
        for investor_wallet in investor_wallets:
            investor_wallet_IT_balance = investor_wallet.IT_balance
            if investor_wallet_IT_balance != 0:
                if investor_wallet_IT_balance >= IT_amount_to_withdraw:
                    IT_amount_to_withdraw_from_wallet = IT_amount_to_withdraw
                else:
                    IT_amount_to_withdraw_from_wallet = investor_wallet_IT_balance

                continue_withdrawing, withdrawn_IT_amount = MatchContract.withdraw_IT(IT_amount_to_withdraw_from_wallet, investor_wallet, reserve)
                IT_amount_to_withdraw -= withdrawn_IT_amount
                print("-----------------")
                print(f"continue withdrawing {continue_withdrawing}")
                print(f"IT amount withdrawn {withdrawn_IT_amount}")
                print(f"IT amount to withdraw {IT_amount_to_withdraw}")
                print("-----------------")
                if not continue_withdrawing or IT_amount_to_withdraw <= 0:
                    break



from fqsdfqsdfqsdfqsd.contracts import RepaymentContract, ReserveContract
from fqsdfqsdfqsdfqsd.oracles import PricingOracle
from fqsdfqsdfqsdfqsd.tokens import RT


class MatchContract:
    @classmethod
    def withdraw_IT(cls, IT_amount, investor_wallet, reserve):
        withdrawn_IT_amount = 0
        if IT_amount > investor_wallet.IT_balance:
            print("-----------------")
            print("you try to withdraw more IT than you have in your wallet")
            print("-----------------")
        else:
            USDC_amount = PricingOracle.IT_to_USDC(IT_amount)
            USDC_amount_to_swap = USDC_amount
            current_IT_price = PricingOracle.IT_price()

            # first swap reserve tokens for IT's
            # update RT holders' wallets IT and USDC balances + burn RT's
            oldest_RTs = reserve.find_oldest_RTs()
            for rt in oldest_RTs:
                RT_USDC_amount = rt.USDC_amount
                if USDC_amount_to_swap - RT_USDC_amount >= 0:
                    print("-----------------")
                    print(f"swapping and burning RT worth {RT_USDC_amount} USDC")
                    print("-----------------")
                    rt.wallet.IT_balance += RT_USDC_amount / current_IT_price
                    investor_wallet.IT_balance -= RT_USDC_amount / current_IT_price
                    RT.del_instance(rt)
                    investor_wallet.USDC_balance += RT_USDC_amount
                    USDC_amount_to_swap -= RT_USDC_amount
                    withdrawn_IT_amount += RT_USDC_amount / current_IT_price
                else:
                    print("-----------------")
                    print(f"swapping and not burning RT worth {USDC_amount_to_swap} USDC")
                    print("-----------------")
                    rt.wallet.IT_balance += USDC_amount_to_swap / current_IT_price
                    investor_wallet.IT_balance -= USDC_amount_to_swap / current_IT_price
                    rt.USDC_amount -= USDC_amount_to_swap
                    investor_wallet.USDC_balance += USDC_amount_to_swap
                    withdrawn_IT_amount += USDC_amount_to_swap / current_IT_price
                    USDC_amount_to_swap = 0
                    break

            # then swap from repayment pool
            if USDC_amount_to_swap != 0:
                if USDC_amount_to_swap <= RepaymentContract.USDC_amount:
                    print("-----------------")
                    print(f"enough capital in the repayment pool; swapping {USDC_amount_to_swap / current_IT_price} ITs")
                    print("-----------------")
                    RepaymentContract.USDC_amount -= USDC_amount_to_swap
                    investor_wallet.IT_balance -= USDC_amount_to_swap / current_IT_price
                    investor_wallet.USDC_balance += USDC_amount_to_swap
                    withdrawn_IT_amount += USDC_amount_to_swap / current_IT_price
                else:
                    print("-----------------")
                    print(f"trying to swap {USDC_amount_to_swap / current_IT_price} ITs, "
                          f"but not enough capital in the repayment pool; swapping {RepaymentContract.USDC_amount / current_IT_price} ITs instead")
                    print("-----------------")
                    investor_wallet.IT_balance -= RepaymentContract.USDC_amount / current_IT_price
                    investor_wallet.USDC_balance += RepaymentContract.USDC_amount
                    withdrawn_IT_amount += RepaymentContract.USDC_amount / current_IT_price
                    RepaymentContract.USDC_amount = 0

                    return False, int(withdrawn_IT_amount)

        return True, int(withdrawn_IT_amount)
